# import time
# from multiprocessing import Pool
#
#
# def f(n):
#     # time.sleep(n)
#     return n
#
#
# pool = Pool(3)
# result = pool.imap(f, (3, 1, 2))
# # result.
# print([res for res in result])
